#include "VirtualPogostick.h"


// Constructor
VirtualPogostick::VirtualPogostick(PlanarLeg *legL, PlanarLeg *legR, bool isFront){
  _legL = legL;
  _legR = legR;
  _is_front = isFront;

  float forwardSpeedInitial = 0.;
  updateDesLimitCycle(forwardSpeedInitial);
  initializeStabilityStateVars();
}

float maxf(float a, float b){
  if(a > b){ return a; }
  return b;
}

float absf(float x){
  if(x < 0){ return -x; }
  return x;
}

void VirtualPogostick::updateDesLimitCycle(float forwardSpeedDes){

  // fore-aft
  _DELTA_XL_0_FP = -_BODY_LEN_HALF -0.0967305*forwardSpeedDes;
  _DELTA_XR_0_FP =  _BODY_LEN_HALF +0.0695445*forwardSpeedDes;
  _DELTA_XL_1_FP = -_BODY_LEN_HALF +0.0967305*forwardSpeedDes;
  _DELTA_XR_1_FP =  _BODY_LEN_HALF -0.0695445*forwardSpeedDes;
  _DXDT_0_FP     = forwardSpeedDes;
  _DXDT_1_FP     = forwardSpeedDes;

  _DELTA_XR_2_FP = -_DELTA_XL_1_FP; 
  _DELTA_XL_2_FP = -_DELTA_XR_1_FP;
  _DELTA_XL_3_FP = -_DELTA_XR_0_FP;
  _DELTA_XR_3_FP = -_DELTA_XL_0_FP;
  _DXDT_2_FP     = forwardSpeedDes;
  _DXDT_3_FP     = forwardSpeedDes;
}

void VirtualPogostick::initializeStabilityStateVars(){
  _touchdown_millis      = 0;
  _liftoff_millis        = 0;
  _touchdown_y           = _VERT_LEG_LEN_GUARD;
  _touchdown_y_other     = _VERT_LEG_LEN_GUARD;
  _touchdown_delta_x     = (_is_front ?  _BODY_LEN_HALF : -_BODY_LEN_HALF);
  _liftoff_y             = _VERT_LEG_LEN_GUARD;
  _liftoff_y_other       = _YR_0_FP;
  _toe_hip_height        = _YR_0_FP;
  _delta_x_est           = (_is_front ?  _BODY_LEN_HALF : -_BODY_LEN_HALF);
  _liftoff_delta_x       = (_is_front ?  _BODY_LEN_HALF : -_BODY_LEN_HALF);
  _liftoff_delta_x_other = (_is_front ? -_BODY_LEN_HALF :  _BODY_LEN_HALF);
  _liftoff_dxdt          = 0.; 
  _dxdt_est              = 0.; 
}


// Needs to be called once per the control loop before the legs are updated
void VirtualPogostick::applyControl(Eigen::Vector3f rpy, Eigen::Vector3f rpyVel, \
                                Eigen::Vector3f rpyOtherPogo, Eigen::Vector3f rpyVelOtherPogo, \
                                float forwardSpeedDes, float yawControlInput, \
                                VirtualPogostick *otherPogo, Spine *spine){


  _misc_log_value = 0.;
  _misc_log_value_2 = 0.;

  // ---------------------------------
  // --------- Preliminaries ---------
  // ---------------------------------
  float tMillis = millis();
  _hybrid_state = _hybrid_state_next;
  updateDesLimitCycle(forwardSpeedDes);

  // used for logging
  _ctl_cmd_l_x = 0.;
  _ctl_cmd_l_y = 0.;
  _ctl_cmd_r_x = 0.;
  _ctl_cmd_r_y = 0.;
  _chart_time_millis = 0.;
  _event_leg_len = _VERT_LEG_LEN_GUARD;
  _dxdt_other_est = otherPogo->getDxdtEst();

  // -----------------------------------------
  // --------- Gather kinematic data ---------
  // -----------------------------------------

  // hip orientation
  _hip_roll = rpy(0); 
  _hip_roll_vel = rpyVel(0);
  _hip_pitch = rpy(1);
  _hip_pitch_vel = rpyVel(1);

  // get the toe in the leveled-hip frame
  Eigen::Vector2f toePosCartHipLeveledFrameL = _legL->getToePosCartesian(_hip_pitch);
  Eigen::Vector2f toePosCartHipLeveledFrameR = _legR->getToePosCartesian(_hip_pitch);
  Eigen::Vector2f toeVelCartHipLeveledFrameL = _legL->getToeVelCartesian(_hip_pitch, \
                                                                         _hip_pitch_vel);
  Eigen::Vector2f toeVelCartHipLeveledFrameR = _legR->getToeVelCartesian(_hip_pitch, \
                                                                         _hip_pitch_vel);
  float dxdtHipFrameLeveled = \
    -(toeVelCartHipLeveledFrameL(0) + toeVelCartHipLeveledFrameR(0))/2.;
  dxdtHipFrameLeveled += (_is_front ? _HORIZ_TRIM_FRONT : _HORIZ_TRIM_REAR);
  float maxHorizontalLegLen = maxf(absf(toePosCartHipLeveledFrameL(0)), \
                                   absf(toePosCartHipLeveledFrameR(0)) );
  float avgToeVelYHipFrameLeveled = \
    (toeVelCartHipLeveledFrameL(1) + toeVelCartHipLeveledFrameR(1))/2.;

  // get the toe in the leveled-body frame
  Eigen::Vector2f framesHip2COM( (_is_front ? _BODY_LEN_HALF : -_BODY_LEN_HALF), 0.);
  Eigen::Vector2f toePosCartBodyLeveledFrameL = toePosCartHipLeveledFrameL + framesHip2COM;
  Eigen::Vector2f toePosCartBodyLeveledFrameR = toePosCartHipLeveledFrameR + framesHip2COM;
  _delta_x_est = (toePosCartBodyLeveledFrameL(0) + toePosCartBodyLeveledFrameR(0))/2.;
  if(_is_front){ _delta_x_est -= _FRONT_SPLAY; } // Account for physical toe offset
  _toe_hip_height = -(toePosCartHipLeveledFrameL(1)+toePosCartHipLeveledFrameR(1))/2.;

  // used for logging
  _pos_toe_hip_leveled_frame_l_x = toePosCartHipLeveledFrameL(0);
  _pos_toe_hip_leveled_frame_l_y = toePosCartHipLeveledFrameL(1);
  _pos_toe_hip_leveled_frame_r_x = toePosCartHipLeveledFrameR(0);
  _pos_toe_hip_leveled_frame_r_y = toePosCartHipLeveledFrameR(1);

  // --------------------------
  // --------- Stance ---------
  // --------------------------
  if(_hybrid_state == PogoState::STANCE){

    // verify that touchdown millis has been set (if starting into STANCE before FLIGHT)
    if(_touchdown_millis == 0){
      _touchdown_millis = tMillis;
      _touchdown_delta_x = _delta_x_est; 
    }

    // measure stance time and estimate dxdt
    _chart_time_millis = tMillis - ((float)_touchdown_millis); 
    float chartTimeSec = _chart_time_millis/1000.;
    if(_chart_time_millis > _DXDT_STANCE_WAIT_MILLIS){
      // we need to wait until the motor's LPF in velocity die out from the retraction (~40ms)
      _dxdt_est = ((1 - _SPEED_EST_COMPL_FILTER_COEF) * dxdtHipFrameLeveled) + \
                       (_SPEED_EST_COMPL_FILTER_COEF * _dxdt_est);
    }

    // check hybrid mode sequence
    bool inDoubleSupport = (otherPogo->getState() == PogoState::STANCE ? true : false );

    // initial transient
    float kdVert = _KD_VERT + _KD_VERT_SPEED * absf(forwardSpeedDes) + \
                              _KD_VERT_SPINE * spine->isActive();
    float stancdCmdY = _STANCE_CMD_Y;
    if(absf(forwardSpeedDes) > 1.1){ 
      // increase applied vertical force at faster speeds as a heuristic to compensate for 
      //   high-current non-linearity in motor current vs. torque relationship.
      if(_is_front){ stancdCmdY -= 2.0; }
      kdVert += 25. * absf(forwardSpeedDes);
    } 
    if(tMillis < _gait_start_millis + _STARTUP_TRANSIENT_MILLIS){
      float lambda = (tMillis - _gait_start_millis)/_STARTUP_TRANSIENT_MILLIS;
      kdVert     = lambda * _KD_VERT      + (1. - lambda) * _KD_VERT_STARTUP;
      stancdCmdY = lambda * _STANCE_CMD_Y + (1. - lambda) * _STANCE_CMD_Y_STARTUP;
    }

    // energize stance vertically to account for unmodeled dissipative forces
    float stanceDurationDesHalfSec = (_FLIGHT_TIME_DES + 2.*_DOUBLE_SUPPORT_TIME_DES)/2.;
    float cmdVertEnergizer = kdVert* (chartTimeSec - stanceDurationDesHalfSec + _T_HALF_OFFSET);

    // horizontal commands
    float yawControlCmd = _KP_YAW * yawControlInput;
    float cmdTemplateX = (_is_front ? \
       _X_ACCEL_2_CMD * (_STANCE_ACCEL_Y/_Y_BAR)*( _BODY_LEN_HALF - _delta_x_est) : \
       _X_ACCEL_2_CMD * (_STANCE_ACCEL_Y/_Y_BAR)*(-_BODY_LEN_HALF - _delta_x_est) );

    // account for spine
    float yCmdSpine = 0.;
    float xCmdSpine = 0.;
    if(spine->isMoving()){
      xCmdSpine = (_is_front ? _X_CMD_SPINE_FRONT : _X_CMD_SPINE_REAR);
      yCmdSpine = (_is_front ? _Y_CMD_SPINE_FRONT : _Y_CMD_SPINE_REAR);
    }

    // double support
    if(inDoubleSupport){
      // no need to generate internal forces in double support
      float deltaXmean = (_delta_x_est + otherPogo->getDeltaX())/2.;
      cmdTemplateX = _X_ACCEL_2_CMD * (_STANCE_ACCEL_Y/_Y_BAR)*(-1.*deltaXmean);
      xCmdSpine = 0.0;
      yCmdSpine = 0.0;
    }
    float xEnergizer = -1. * _X_POS_DAMPING * forwardSpeedDes;
    float cmdXL = cmdTemplateX + yawControlCmd + xEnergizer + xCmdSpine;
    float cmdXR = cmdTemplateX - yawControlCmd + xEnergizer + xCmdSpine;

    // vertical commands (leg should be pushing downwards, not upwards!)
    float cmdYDiffRoll = (_KP_ROLL * _hip_roll) + (_KD_ROLL * _hip_roll_vel);
    float cmdYL = stancdCmdY + cmdYDiffRoll + cmdVertEnergizer + yCmdSpine;
    float cmdYR = stancdCmdY - cmdYDiffRoll + cmdVertEnergizer + yCmdSpine;

    // send commands to legs
    _legL->setOpenLoopCmdCartesian(cmdXL, cmdYL, _hip_pitch);
    _legR->setOpenLoopCmdCartesian(cmdXR, cmdYR, _hip_pitch);

    // grab data from other pogostick after it initiates double support
    float doubleSupportChartTimeSec = _DOUBLE_SUPPORT_TIME_DES;
    float doubleSupportY0 = _YR_1_FP;
    float doubleSupportOtherY0 = _YL_1_FP;
    uint32_t otherPogoTouchdownTime = otherPogo->getTouchdownMillis();
    if(otherPogoTouchdownTime > _touchdown_millis){ // double support
      doubleSupportChartTimeSec = (tMillis - ((float)otherPogoTouchdownTime))/1000.;
      doubleSupportY0 = otherPogo->getTouchdownHipHeightOther();
      doubleSupportOtherY0 = otherPogo->getTouchdownHipHeight();
    }

    // form the virtual stance leg length for liftoff
    float legLengthForLiftoff = _VERT_LEG_LEN_GUARD + \
      _C_TAU_2 * (doubleSupportChartTimeSec - _DOUBLE_SUPPORT_TIME_DES ) + \
      _C_R_21  * (doubleSupportY0           - _YR_1_FP) + \
      _C_L_21  * (doubleSupportOtherY0      - _YL_1_FP);

    // used for logging
    _ctl_cmd_l_x = cmdXL;
    _ctl_cmd_l_y = cmdYL;
    _ctl_cmd_r_x = cmdXR;
    _ctl_cmd_r_y = cmdYR;
    _event_leg_len = legLengthForLiftoff;

    // transition from STANCE to FLIGHT if necessary
    bool liftoffBool = false;
    if(_toe_hip_height > legLengthForLiftoff && _chart_time_millis > _MIN_STANCE_MILLIS && \
       avgToeVelYHipFrameLeveled < _DYDT_TOE_VEL_LO_THRESHOLD && inDoubleSupport){
      liftoffBool = true; // normal liftoff
    }
    if(_chart_time_millis > _MIN_STANCE_MILLIS && maxHorizontalLegLen > _MAX_LEG_LEN_HORIZONTAL){
      liftoffBool = true; // early liftoff to prevent high-speed sumble
    }
    if(_chart_time_millis > _MIN_STANCE_MILLIS && _toe_hip_height > _MAX_LEG_LEN_STANCE){
      liftoffBool = true;
    }
    if(_chart_time_millis > _MAX_STANCE_MILLIS){
      liftoffBool = true;
    }

    if(liftoffBool){
      // set hybrid state to FLIGHT and reset STANCE variables
      _hybrid_state_next = PogoState::FLIGHT;
      _prev_chart_duration = _chart_time_millis;
      _liftoff_millis = tMillis;
		  _liftoff_y = _toe_hip_height;
      _liftoff_y_other = otherPogo->getStanceHipHeight();
      _liftoff_delta_x = _delta_x_est;
      _liftoff_delta_x_other = otherPogo->getDeltaX();
      _liftoff_dxdt = _dxdt_est;
      _prev_chart_duration = _chart_time_millis;
    }
  }
  // --------------------------
  // --------- Flight ---------
  // --------------------------
  else if(_hybrid_state == PogoState::FLIGHT){

    // measure flight time
    _chart_time_millis = tMillis - ((float)_liftoff_millis);
    float flightChartTimeSec = _chart_time_millis/1000.; 

    // form the desired flight vertical leg length/vel
    float flightChartTimeSecEff = flightChartTimeSec;
    float legLengthForTouchdown = _VERT_LEG_LEN_GUARD - _TOUCHDOWN_TOE_ERROR_THRESH + \
        _C_TAU_1 * (flightChartTimeSecEff - _FLIGHT_TIME_DES) + \
        _C_L_10  * (_liftoff_y            - _YL_0_FP        ) + \
        _C_R_10  * (_liftoff_y_other      - _YR_0_FP        );
    float yToeDesHipFrameLeveled = -legLengthForTouchdown;
    float velVertProtraction = -_C_TAU_1;
    if(yToeDesHipFrameLeveled > _Y_MAX_TOE_FLIGHT){
      yToeDesHipFrameLeveled = _Y_MAX_TOE_FLIGHT;
      velVertProtraction = 0.0;
    }else if(yToeDesHipFrameLeveled < _Y_MIN_TOE_FLIGHT){
      yToeDesHipFrameLeveled = _Y_MIN_TOE_FLIGHT;
      velVertProtraction = 0.0;
    }

    // desired horizontal leg len
    float velHorzProtraction = 0.;
    float deltaXDes;
    float deltaXOther = otherPogo->getDeltaX();
    if(!_is_front){
      float errorTerm = \
        _K_DELTA_XL_10 * (_liftoff_delta_x       - _DELTA_XL_0_FP  ) + \
        _K_DELTA_XR_10 * (_liftoff_delta_x_other - _DELTA_XR_0_FP  ) + \
        _K_DELTA_XR_11 * (deltaXOther            - _DELTA_XR_1_FP  ) + \
        _K_DXDT_10     * (_liftoff_dxdt          - _DXDT_0_FP      ) + \
        _K_DXDT_11     * (_dxdt_other_est        - _DXDT_1_FP      );
      deltaXDes = _DELTA_XL_1_FP + errorTerm;
    }else{
      float errorTerm = \
        _K_DELTA_XR_32 * (_liftoff_delta_x       - _DELTA_XR_2_FP  ) + \
        _K_DELTA_XL_32 * (_liftoff_delta_x_other - _DELTA_XL_2_FP  ) + \
        _K_DELTA_XL_33 * (deltaXOther            - _DELTA_XL_3_FP  ) + \
        _K_DXDT_32     * (_liftoff_dxdt          - _DXDT_2_FP      ) + \
        _K_DXDT_33     * (_dxdt_other_est        - _DXDT_3_FP      );
      deltaXDes = _DELTA_XR_3_FP  + errorTerm;
    }
    deltaXDes += _TD_DETECTION_DELAY_SEC * forwardSpeedDes; // account for TD delay
    if(_is_front){ deltaXDes += _FRONT_SPLAY; } // Account for physical toe offset

    // convert des x pos to local coordinates
    float xToeDesHipFrameLeveled = deltaXDes - framesHip2COM(0);
    if(xToeDesHipFrameLeveled > _MAX_LEG_LEN_HORIZONTAL){
      xToeDesHipFrameLeveled = _MAX_LEG_LEN_HORIZONTAL;
    }else if(xToeDesHipFrameLeveled < -_MAX_LEG_LEN_HORIZONTAL){
      xToeDesHipFrameLeveled = -_MAX_LEG_LEN_HORIZONTAL;
    }

    // send leg commands
    _legL->setCartesianGains(_KP_FLIGHT_TOE, _KD_FLIGHT_TOE,\
                             _KP_FLIGHT_TOE, _KD_FLIGHT_TOE);
    _legR->setCartesianGains(_KP_FLIGHT_TOE, _KD_FLIGHT_TOE,\
                             _KP_FLIGHT_TOE, _KD_FLIGHT_TOE);
    _legL->setPosCmdToeCartesian(xToeDesHipFrameLeveled, yToeDesHipFrameLeveled, \
                                     velHorzProtraction, velVertProtraction, \
                                             _hip_pitch, _hip_pitch_vel);
    _legR->setPosCmdToeCartesian(xToeDesHipFrameLeveled, yToeDesHipFrameLeveled, \
                                     velHorzProtraction, velVertProtraction, \
                                             _hip_pitch, _hip_pitch_vel);

    // used for logging
    _ctl_cmd_l_x = xToeDesHipFrameLeveled;
    _ctl_cmd_l_y = yToeDesHipFrameLeveled;
    _ctl_cmd_r_x = xToeDesHipFrameLeveled;
    _ctl_cmd_r_y = yToeDesHipFrameLeveled;
    _event_leg_len = legLengthForTouchdown;

    // transition to stance if necessary (transient time allows the leg to get moving downwards)
    float toeYErrorL = yToeDesHipFrameLeveled - _pos_toe_hip_leveled_frame_l_y;
    float toeYErrorR = yToeDesHipFrameLeveled - _pos_toe_hip_leveled_frame_r_y;
    float toeYErrorAvg = (toeYErrorL + toeYErrorR)/2.;
    if(_chart_time_millis > _FLIGHT_TIME_MIN_MILLIS){
      if( toeYErrorAvg < _TOUCHDOWN_TOE_ERROR_THRESH ){
        // change hybrid state and perform bookkeeping
        _hybrid_state_next = PogoState::STANCE;
        _touchdown_millis = tMillis;
        _touchdown_delta_x = _delta_x_est;
		    _touchdown_y = _toe_hip_height;
        _touchdown_y_other = otherPogo->getStanceHipHeight();
        _prev_chart_duration = _chart_time_millis;
        _dxdt_est = _liftoff_dxdt; // seed the filter with the other dxdt est
      }
    }
  }
  // ------------------------
  // --------- Rest ---------
  // ------------------------
  else if(_hybrid_state == PogoState::REST || _hybrid_state == PogoState::STANDING_UP){
    if(_hybrid_state == PogoState::STANDING_UP){
      _t_stand_millis = tMillis;
    }
    // just sit tight at desired position (don't use IMU for this)
    float xToeDesBodyFrameLeveled = (_is_front ? _BODY_LEN_HALF: -_BODY_LEN_HALF);
    float xToeDesHipFrameLeveled = xToeDesBodyFrameLeveled - framesHip2COM(0);
    float yToeDesHipFrameLeveled = -1.*(_VERT_LEG_LEN_GUARD + _REST_LEG_LEN_Y_OFFSET); 
    _legL->setCartesianGains(_KP_REST, _KD_REST, _KP_REST, _KD_REST);
    _legR->setCartesianGains(_KP_REST, _KD_REST, _KP_REST, _KD_REST);
    _legL->setPosCmdToeCartesian(xToeDesHipFrameLeveled, yToeDesHipFrameLeveled);
    _legR->setPosCmdToeCartesian(xToeDesHipFrameLeveled, yToeDesHipFrameLeveled);
    // make sure this was set properly
    zeroGaitVariables();
    initializeStabilityStateVars();
    _legL->activateSingularityAvoidance();
		_legR->activateSingularityAvoidance();
    // if just starting up, stand up slowly
    float lambda = (tMillis - _t_stand_millis)/_STAND_DURATION_MILLIS;
    lambda = (lambda > 1. ? 1. : lambda);
    lambda = (lambda < 0. ? 0. : lambda);
		_legL->setMaxMotCmd(lambda * _MAX_LEG_CMDS);
		_legR->setMaxMotCmd(lambda * _MAX_LEG_CMDS);
    if(_hybrid_state == PogoState::STANDING_UP){
      Serial1 << "STANDING UP in hopper " << _is_front << " at time = " << tMillis << "\r\n";
      _hybrid_state_next = PogoState::REST;
      _t_stand_millis = tMillis; 
    }
  }
  // -----------------------------
  // --------- Power off ---------
  // -----------------------------
  else if(_hybrid_state == PogoState::POWER_OFF){
    zeroGaitVariables();
    _legL->deactivateSingularityAvoidance();
    _legR->deactivateSingularityAvoidance();
    _legL->setMaxMotCmd(0.);
    _legR->setMaxMotCmd(0.);
    _legL->setOpenLoopCmdMotors(0.,0.);
    _legR->setOpenLoopCmdMotors(0.,0.);
  }

  // used for logging
  Eigen::Vector2f cmdL = _legL->getMotorCommand();
  Eigen::Vector2f cmdR = _legR->getMotorCommand();
  _mot_cmd_ll = cmdL(0); 
  _mot_cmd_lr = cmdL(1); 
  _mot_cmd_rl = cmdR(0); 
  _mot_cmd_rr = cmdR(1); 
 
}

// this function makes sure user input doesn't screw up stance variables
void VirtualPogostick::zeroGaitVariables(){
  _touchdown_millis = 0;
  _prev_chart_duration = 0.;
}

unsigned VirtualPogostick::stateToInt(PogoState state){
  switch(state){
    case PogoState::STANCE:
      return 1;
    case PogoState::FLIGHT:
      return 2;
    case PogoState::REST:
      return 3;
    case PogoState::POWER_OFF:
      return 4;
    case PogoState::STANDING_UP:
      return 5;
    default:
      return 0;
  }
}



void VirtualPogostick::beginGait(){
  this->setState(PogoState::STANCE);
  _gait_start_millis = millis();
}
