#ifndef __VIRTUAL_POGOSTICK_H__
#define __VIRTUAL_POGOSTICK_H__

#include <Arduino.h>
#include <Parallel5Bar.h>  
#include <MathApproximations.h>
#include <Mapping.h>
#include <Controls.h>
#include <RC.h> 
#include "Spine.h"

extern RC *rc;

enum class PogoState {STANCE, FLIGHT, REST, POWER_OFF, STANDING_UP};
enum class PhaseRelation {GOOD, BEHIND};

class VirtualPogostick {

  // the hipLeveledFrame is the hip frame rotated to be level with the world (flat) ground
  // the bodyLeveledFrame is the body frame rotated to be level with the world (flat) ground
  //   using the estimated BODY PITCH

 public:

    // Constructor
    VirtualPogostick(PlanarLeg *legL, PlanarLeg *legR, bool isFront);

    // Needs to be called once per the control loop before the legs are updated
    void applyControl(Eigen::Vector3f rpy,          Eigen::Vector3f rpyVel, \
                      Eigen::Vector3f rpyOtherPogo, Eigen::Vector3f rpyVelOtherPogo, \
                      float forwardSpeedDes, float yawControlInput, \
                      VirtualPogostick *otherPogo, Spine *spine);

    // used to transition hybridState
    void setState(PogoState stateToSet){
      _hybrid_state = stateToSet; 
      _hybrid_state_next = stateToSet;
    }
    PogoState getState(){return _hybrid_state;}
    unsigned getStateInt(){return stateToInt(_hybrid_state);}
    float getTouchdownTime(){return _touchdown_millis;}
    bool hasGaitStarted(){return _hybrid_state == PogoState::STANCE \
                              || _hybrid_state == PogoState::FLIGHT;}
    void beginGait();
    
    // used for inter-pogo communication on the guard
    float getStanceHipHeight(){return _toe_hip_height;}
    uint32_t getTouchdownMillis(){return _touchdown_millis;}
    float getTouchdownHipHeight(){return _touchdown_y;}
    float getTouchdownHipHeightOther(){return _touchdown_y_other;}
    float getLiftoffHipHeight(){return _liftoff_y;}
    float getLiftoffHipHeightOther(){return _liftoff_y_other;}
    float getDeltaX(){return _delta_x_est;}
    float getDxdtEst(){return _dxdt_est;}
    float getDxdtEstOther(){return _dxdt_other_est;}
    float getLiftoffDeltaXOther(){return _liftoff_delta_x_other;}

    // used for logging
    float getToeXL(){return _pos_toe_hip_leveled_frame_l_x;}
    float getToeYL(){return _pos_toe_hip_leveled_frame_l_y;}
    float getToeXR(){return _pos_toe_hip_leveled_frame_r_x;}
    float getToeYR(){return _pos_toe_hip_leveled_frame_r_y;}
    float getCmdXL(){return _ctl_cmd_l_x;}
    float getCmdYL(){return _ctl_cmd_l_y;}
    float getCmdXR(){return _ctl_cmd_r_x;}
    float getCmdYR(){return _ctl_cmd_r_y;}
    float getCmdLL(){return _mot_cmd_ll;}
    float getCmdLR(){return _mot_cmd_lr;}
    float getCmdRL(){return _mot_cmd_rl;}
    float getCmdRR(){return _mot_cmd_rr;}
    float getHipPitch(){return _hip_pitch;}
    float getHipRoll(){return _hip_roll;}
    float getChartTimeMillis(){return _chart_time_millis;}
    float getMiscLogValue(){return _misc_log_value;}
    float getEventLegLen(){return _event_leg_len;}
    float getMiscLogValue2(){return _misc_log_value_2;}
    float getPrevChartDuration(){return _prev_chart_duration; }
    bool isTransientActive(){return _starting_transient_active; }
    uint32_t getGaitStartMillis(){ return _gait_start_millis; }

  protected: 
    // Note: stance cmds are from the hip's toe frame of reference in the sense that 
    //  negative y commands push the leg down, generating a positive y GRF on the mass center

    // general/kinematics 
    const float _BODY_LEN         =  0.47;  
    const float _BODY_LEN_HALF    = _BODY_LEN/2.;
    const float _MAX_LEG_CMDS     =  1.0;
    const float _MAX_LEG_LEN_HORIZONTAL = 0.16;

    // --- vertical hopping template parameters ---
    // core parameters
    const float _VERT_LEG_LEN_GUARD          =  0.22;
    const float _FLIGHT_TIME_DES             =  0.15;   // (sec)
    const float _DOUBLE_SUPPORT_TIME_DES     =  0.02733; // (sec)
    // energy injection (kd used to counter the friction in legs)
    const float _STANCE_ACCEL_Y              =  8.5;
    const float _STANCE_CMD_Y                = -9.0;
    const float _KD_VERT                     = -20.;
    const float _KD_VERT_SPEED               = -25.;
    const float _T_HALF_OFFSET               =   0.014;
    // startup transient
    const float _STARTUP_TRANSIENT_MILLIS    = 2000.;
    const float _STANCE_CMD_Y_STARTUP        =  -8.0;
    const float _KD_VERT_STARTUP             =  -0.;
    // additional stance parameters
    const float _FRONT_SPLAY                 =  0.03; // To account for toe's physical offset
    const float _DYDT_TOE_VEL_LO_THRESHOLD   =  -0.2; // LO guard component
    const float _MIN_STANCE_MILLIS           =   50.; // prevent early LO
    const float _MAX_STANCE_MILLIS           =  210.; // for initial transient
    const float _MAX_LEG_LEN_STANCE          =  0.24;
    const float _KP_ROLL                     =  40.0;
    const float _KD_ROLL                     =   2.0;
    const float _KP_YAW                      =   1.0;
    // additional flight parameters
    const float _Y_MAX_TOE_FLIGHT            =  -0.15; // saturates leg retraction
    const float _Y_MIN_TOE_FLIGHT            =  -0.25; // saturates leg retraction
    const float _KP_FLIGHT_TOE               =  75.0;
    const float _KD_FLIGHT_TOE               =   3.0;
    const float _TOUCHDOWN_TOE_ERROR_THRESH  =  -0.02;
    const float _FLIGHT_TIME_MIN_MILLIS      = 100.;

    // --- fore/aft template parameters ---
    // core parameters
    const float _X_ACCEL_2_CMD               =  -0.50;
    const float _Y_BAR                       =   0.21; // avg COM height in stance under Vicon
    // energy injection
    const float _X_POS_DAMPING               =   0.3;
    // additional parameters
    const float _SPEED_EST_COMPL_FILTER_COEF =   0.97;
    const float _DXDT_STANCE_WAIT_MILLIS     =  20.0;
    const float _HORIZ_TRIM_FRONT            =  -0.08;  // offset in how the legs measure speed to agree w/ Vicon data
    const float _HORIZ_TRIM_REAR             =   0.04;  // offset in how the legs measure speed to agree w/ Vicon data
    const float _TD_DETECTION_DELAY_SEC      =   0.00;

    // --- spine parameters ---
    const float _X_CMD_SPINE_FRONT           = 0.*  -0.5;
    const float _X_CMD_SPINE_REAR            =   -1.0;
    const float _Y_CMD_SPINE_FRONT           =   -6.0;
    const float _Y_CMD_SPINE_REAR            = 0. *  -3.0;
    const float _KD_VERT_SPINE               = 0. * -15.;

    // --- limit cycle variables ---
    // vertical 
    const float _YL_0_FP = _VERT_LEG_LEN_GUARD;
    const float _YR_0_FP = 0.203448;
    const float _YL_1_FP = _VERT_LEG_LEN_GUARD;
    const float _YR_1_FP = _YR_0_FP;

    // fore-aft
    float _DELTA_XL_0_FP, _DELTA_XR_0_FP, _DELTA_XL_1_FP, _DELTA_XR_1_FP;
    float _DELTA_XR_2_FP, _DELTA_XL_2_FP, _DELTA_XL_3_FP, _DELTA_XR_3_FP;
    float _DXDT_0_FP, _DXDT_1_FP, _DXDT_2_FP, _DXDT_3_FP;

    // --- vertical stability parameters ---
     // vertical flight leg protraction
    const float _C_L_10  =  0.544302;
    const float _C_R_10  = -0.081429;
    const float _C_TAU_1 =  0.298966;
    // vertical stance leg protraction
    const float _C_L_21  =  0.426694;
    const float _C_R_21  =  0.;
    const float _C_TAU_2 = -0.313940;

    // --- fore/aft stability parameters ---
    // rear hip
    const float _K_DXDT_10     =  0.;
    const float _K_DXDT_11     =  0.206528;
    const float _K_DELTA_XL_10 =  0.;
    const float _K_DELTA_XR_10 = -0.126227;
    const float _K_DELTA_XR_11 =  0.;
    // front hip
    const float _K_DXDT_32     =  0.;
    const float _K_DXDT_33     =  0.206528;
    const float _K_DELTA_XR_32 =  0.;
    const float _K_DELTA_XL_32 = -0.126227;
    const float _K_DELTA_XL_33 =  0.;

    // RESTING gains
    const float _KP_REST               = 150.0;
    const float _KD_REST               =   1.0;
    const float _REST_LEG_LEN_Y_OFFSET = -0.04;
    const float _STAND_DURATION_MILLIS = 2000; // gradually increase motor power over this time
    const float _K_HIP_PITCH_LP        = 0.0025;
    const uint32_t _TRANSIENT_TIME_MILLIS = 2000;

    // used for logging  
    float _pos_toe_hip_leveled_frame_l_x = 0.;
    float _pos_toe_hip_leveled_frame_l_y = 0.;
    float _pos_toe_hip_leveled_frame_r_x = 0.;
    float _pos_toe_hip_leveled_frame_r_y = 0.;
    float _ctl_cmd_l_x = 0.;
    float _ctl_cmd_l_y = 0.;
    float _ctl_cmd_r_x = 0.;
    float _ctl_cmd_r_y = 0.;
    float _mot_cmd_ll  = 0.;
    float _mot_cmd_lr  = 0.;
    float _mot_cmd_rl  = 0.;
    float _mot_cmd_rr  = 0.;
    float _hip_pitch   = 0.;
    float _hip_pitch_vel = 0.;
    float _hip_roll      = 0.;
    float _hip_roll_vel  = 0.;
    float _chart_time_millis = 0.;
    float _misc_log_value    = 0.;
    float _misc_log_value_2  = 0.;
    float _event_leg_len     = 0.;
    float _prev_chart_duration = 0.;
    uint32_t _gait_start_millis = 0;
    bool _starting_transient_active = false;

    // State
    bool _is_front; // true if front, false if rear
    PlanarLeg *_legL, *_legR; // left and right legs
    PogoState  _hybrid_state          = PogoState::POWER_OFF;
    PogoState  _hybrid_state_next     = PogoState::POWER_OFF;
    float      _t_stand_millis        = 0.;
    // Used for stability controller
    uint32_t   _touchdown_millis      = 0;
    uint32_t   _liftoff_millis        = 0;
    float      _touchdown_y           = 0.;
    float      _touchdown_y_other     = 0.;
    float      _touchdown_delta_x     = 0.;
    float      _liftoff_y             = 0.;
    float      _liftoff_y_other       = 0.;
    float      _toe_hip_height        = 0.;
    float      _delta_x_est           = 0.; 
    float      _liftoff_delta_x       = 0.;
    float      _liftoff_delta_x_other = 0.; 
    float      _liftoff_dxdt          = 0.; 
    float      _dxdt_est              = 0.; 
    float      _dxdt_other_est        = 0.;
    float      _toe_rad_vel_avg_lpf   = 0.;

    // Private member functions
    void zeroGaitVariables(); // used so user input doesn't screw up stance variables
    unsigned stateToInt(PogoState state); 
    Eigen::Matrix2f rotationMatrix(float angleRad);
    void updateDesLimitCycle(float forwardSpeed);
    void initializeStabilityStateVars();
    
 };
#endif
